"""
The Davies-Bouldin (DB) Cluster Validity Index.
"""

# Custom imports
import numpy as np

# Local imports
from . import _base


# DB object definition
class DB(_base.CVI):
    """
    The stateful information of the Davies-Bouldin (DB) Cluster Validity Index.

    References
    ----------
    1. D. L. Davies and D. W. Bouldin, "A cluster separation measure," IEEE Transaction on Pattern Analysis and Machine Intelligence, vol. 1, no. 2, pp. 224-227, Feb. 1979.
    2. M. Moshtaghi, J. C. Bezdek, S. M. Erfani, C. Leckie, and J. Bailey, "Online Cluster Validity Indices for Streaming Data," ArXiv e-prints, 2018, arXiv:1801.02937v1 [stat.ML]. [Online].
    3. M. Moshtaghi, J. C. Bezdek, S. M. Erfani, C. Leckie, J. Bailey, "Online cluster validity indices for performance monitoring of streaming data clustering," Int. J. Intell. Syst., pp. 1-23, 2018.
    """

    def __init__(self):
        """
        Davies-Bouldin (DB) initialization routine.
        """

        # Run the base initialization
        super().__init__()

        # CH-specific initialization
        self.mu = np.zeros([0])     # dim
        self.R = np.zeros([0, 0])   # n_clusters x dim
        self.D = np.zeros([0, 0])   # n_clusters x n_clusters
        self.S = []                 # dim

        return

    @_base.add_docs(_base.setup_doc)
    def setup(self, sample: np.ndarray) -> None:
        """
        Davies-Bouldin (DB) setup routine.
        """

        # Run the generic setup routine
        super().setup(sample)

        # CH-specific setup
        self.mu = sample

        return

    @_base.add_docs(_base.param_inc_doc)
    def param_inc(self, sample: np.ndarray, label: int) -> None:
        """
        Incremental parameter update for the Davies-Bouldin (DB) CVI.
        """

        # Get the internal label corresponding to the provided label
        i_label = self.label_map.get_internal_label(label)

        # Increment the local number of samples count
        n_samples_new = self.n_samples + 1

        # Check if the module has been setup, then set the mu accordingly
        if self.n_samples == 0:
            self.setup(sample)
        else:
            self.mu = (1 - 1/n_samples_new) * self.mu + (1/n_samples_new) * sample

        # IF NEW CLUSTER LABEL
        # Correct for python 0-indexing
        if i_label > self.n_clusters - 1:
            n_new = 1
            v_new = sample
            CP_new = 0.0
            G_new = np.zeros(self.dim)
            S_new = 0.0
            if self.n_clusters == 0:
                D_new = np.zeros((1, 1))
            else:
                D_new = np.zeros((self.n_clusters + 1, self.n_clusters + 1))
                D_new[0:self.n_clusters, 0:self.n_clusters] = self.D
                d_column_new = np.zeros(self.n_clusters + 1)
                for jx in range(self.n_clusters):
                    d_column_new[jx] = (
                        np.sum((v_new - self.v[jx, :]) ** 2)
                    )
                D_new[i_label, :] = d_column_new
                # D_new[:, i_label] = np.transpose(d_column_new)
                D_new[:, i_label] = d_column_new

            # Update 1-D parameters with list appends
            self.n_clusters += 1
            self.n.append(n_new)
            self.CP.append(CP_new)
            self.S.append(S_new)

            # Update 2-D parameters with numpy vstacks
            self.v = np.vstack([self.v, v_new])
            self.G = np.vstack([self.G, G_new])
            self.D = D_new

        # ELSE OLD CLUSTER LABEL
        else:
            n_new = self.n[i_label] + 1
            v_new = (
                (1 - 1 / n_new) * self.v[i_label, :]
                + (1 / n_new) * sample
            )
            delta_v = self.v[i_label, :] - v_new
            diff_x_v = sample - v_new
            CP_new = (
                self.CP[i_label]
                + np.inner(diff_x_v, diff_x_v)
                + self.n[i_label] * np.inner(delta_v, delta_v)
                + 2 * np.inner(delta_v, self.G[i_label, :])
            )
            G_new = (
                self.G[i_label, :]
                + diff_x_v
                + self.n[i_label] * delta_v
            )
            S_new = CP_new / n_new
            d_column_new = np.zeros(self.n_clusters)
            for jx in range(self.n_clusters):
                # Skip the current i_label index
                if jx == i_label:
                    continue
                d_column_new[jx] = (
                    np.sum((v_new - self.v[jx, :]) ** 2)
                )

            # Update parameters
            self.n[i_label] = n_new
            self.v[i_label, :] = v_new
            self.CP[i_label] = CP_new
            self.G[i_label, :] = G_new
            self.S[i_label] = S_new
            self.D[i_label, :] = d_column_new
            # self.D[:, i_label] = np.tranpose(d_column_new)
            self.D[:, i_label] = d_column_new

        # Update the parameters that do not depend on label novelty
        self.n_samples = n_samples_new

        return

    @_base.add_docs(_base.param_batch_doc)
    def param_batch(self, data: np.ndarray, labels: np.ndarray) -> None:
        """
        Batch parameter update for the Davies-Bouldin (DB) CVI.
        """

        # Setup the CVI for batch mode
        super().setup_batch(data)

        # Take the average across all samples, but cast to 1-D vector
        self.mu = np.mean(data, axis=0)
        u = np.unique(labels)
        self.n_clusters = u.size
        # self.n = np.zeros(self.n_clusters, dtype=int)
        self.n = [0 for _ in range(self.n_clusters)]
        self.v = np.zeros((self.n_clusters, self.dim))
        # self.CP = np.zeros(self.n_clusters)
        self.CP = [0 for _ in range(self.n_clusters)]
        self.G = np.zeros((self.n_clusters, self.dim))
        self.D = np.zeros((self.n_clusters, self.n_clusters))
        # self.S = np.zeros(self.n_clusters)
        self.S = [0 for _ in range(self.n_clusters)]

        for ix in range(self.n_clusters):
            # subset_indices = lambda x: labels[x] == ix
            subset_indices = (
                [x for x in range(len(labels)) if labels[x] == ix]
            )
            subset = data[subset_indices, :]
            self.n[ix] = subset.shape[0]
            self.v[ix, :] = np.mean(subset, axis=0)
            diff_x_v = subset - self.v[ix, :] * np.ones((self.n[ix], 1))
            self.CP[ix] = np.sum(diff_x_v ** 2)
            self.S[ix] = self.CP[ix] / self.n[ix]

        for ix in range(self.n_clusters - 1):
            for jx in range(ix + 1, self.n_clusters):
                self.D[ix, jx] = (
                    np.sum((self.v[ix, :] - self.v[jx, :]) ** 2)
                )

        self.D = self.D + np.transpose(self.D)

        return

    @_base.add_docs(_base.evaluate_doc)
    def evaluate(self) -> None:
        """
        Criterion value evaluation method for the Davies-Bouldin (DB) CVI.
        """
        self.R = np.zeros((self.n_clusters, self.n_clusters))

        if self.n_clusters > 2:
            for ix in range(self.n_clusters - 1):
                for jx in range(ix + 1, self.n_clusters):
                    self.R[jx, ix] = (
                        (self.S[ix] + self.S[jx]) / self.D[jx, ix]
                    )
            self.R = self.R + np.transpose(self.R)
            self.criterion_value = (
                np.sum(np.max(self.R, axis=0)) / self.n_clusters
            )
        else:
            self.criterion_value = 0.0

        return



"""
The :mod:`modules` gathers all CVI implementations.
"""

from ._base import (
    LabelMap,
    CVI,
    add_docs,
    param_inc_doc,
    param_batch_doc,
)

from .CH import CH
from .cSIL import cSIL
from .DB import DB

__all__ = [
    "LabelMap",
    "CVI",
    "add_docs",
    "param_inc_doc",
    "param_batch_doc",
    "CH",
    "cSIL",
    "DB",
]



"""
The Centroid-based Silhouette (cSIL) Cluster Validity Index.
"""

# Custom imports
import numpy as np

# Local imports
from . import _base


# cSIL object definition
class cSIL(_base.CVI):
    """
    The stateful information of the Centroid-based Silhouette (cSIL) Cluster Validity Index.

    References
    ----------
    1. L. E. Brito da Silva, N. M. Melton, and D. C. Wunsch II, "Incremental Cluster Validity Indices for Hard Partitions: Extensions  and  Comparative Study," ArXiv  e-prints, Feb 2019, arXiv:1902.06711v1 [cs.LG].
    2. P. J. Rousseeuw, "Silhouettes: A graphical aid to the interpretation and validation of cluster analysis," Journal of Computational and Applied Mathematics, vol. 20, pp. 53–65, 1987.
    3. M. Rawashdeh and A. Ralescu, "Center-wise intra-inter silhouettes," in Scalable Uncertainty Management, E. Hüllermeier, S. Link, T. Fober et al., Eds. Berlin, Heidelberg: Springer, 2012, pp. 406–419.
    """

    def __init__(self):
        """
        Centroid-based Silhouette (cSIL) initialization routine.
        """

        # Run the base initialization
        super().__init__()

        # cSIL-specific initialization
        self.S = np.empty([0, 0])   # n_clusters x dim
        self.sil_coefs = []         # dim

        return

    @_base.add_docs(_base.setup_doc)
    def setup(self, sample: np.ndarray) -> None:
        """
        Centroid-based Silhouette (cSIL) setup routine.
        """

        # Run the generic setup routine
        super().setup(sample)

        return

    @_base.add_docs(_base.param_inc_doc)
    def param_inc(self, sample: np.ndarray, label: int) -> None:
        """
        Incremental parameter update for the Centroid-based Silhouette (cSIL) CVI.
        """

        # Get the internal label corresponding to the provided label
        i_label = self.label_map.get_internal_label(label)

        # Increment the local number of samples count
        n_samples_new = self.n_samples + 1

        # Check if the module has been setup, then set the mu accordingly
        if self.n_samples == 0:
            self.setup(sample)

        # IF NEW CLUSTER LABEL
        # Correct for python 0-indexing
        if i_label > self.n_clusters - 1:
            n_new = 1
            v_new = sample
            CP_new = np.inner(sample, sample)
            G_new = sample

            # Compute S_new
            if self.n_clusters == 0:
                S_new = np.zeros([1, 1])
            else:
                S_new = np.zeros((self.n_clusters + 1, self.n_clusters + 1))
                S_new[0:self.n_clusters, 0:self.n_clusters] = self.S
                S_row_new = np.zeros(self.n_clusters + 1)
                S_col_new = np.zeros(self.n_clusters + 1)
                for cl in range(self.n_clusters):
                    # Column "bmu_temp - D_new"
                    C = (
                        CP_new
                        + np.inner(self.v[cl, :], self.v[cl, :])
                        - np.inner(G_new, self.v[cl, :])
                    )
                    S_col_new[cl] = C
                    C = (
                        self.CP[cl]
                        + self.n[cl] * np.inner(v_new, v_new)
                        - 2 * np.inner(self.G[cl, :], v_new)
                    )
                    S_row_new[cl] = C / self.n[cl]
                # Column "ind_minus" - F
                S_col_new[i_label] = 0
                S_row_new[i_label] = S_col_new[i_label]
                S_new[i_label, :] = S_col_new
                S_new[:, i_label] = S_row_new

            # Update 1-D parameters with list appends
            self.n_clusters += 1
            self.n.append(n_new)
            self.CP.append(CP_new)

            # Update 2-D parameters with numpy vstacks
            self.v = np.vstack([self.v, v_new])
            self.G = np.vstack([self.G, G_new])
            self.S = S_new

        # ELSE OLD CLUSTER LABEL
        else:
            n_new = self.n[i_label] + 1
            v_new = (
                (1 - 1 / n_new) * self.v[i_label, :]
                + (1 / n_new) * sample
            )
            # delta_v = self.v[i_label, :] - v_new
            # diff_x_v = sample - v_new
            CP_new = (
                self.CP[i_label]
                + np.inner(sample, sample)
            )
            G_new = (
                self.G[i_label, :]
                + sample
            )
            # Compute S_new
            S_row_new = np.zeros(self.n_clusters)
            S_col_new = np.zeros(self.n_clusters)
            for cl in range(self.n_clusters):
                # Skip the i_label iteration
                if cl == i_label:
                    continue
                # Column "bmu_temp" - D_new
                diff_x_v = sample - self.v[cl, :]
                C = (
                    self.CP[i_label]
                    + np.inner(diff_x_v, diff_x_v)
                    + self.n[i_label] * np.inner(self.v[cl, :], self.v[cl, :])
                    - 2 * np.inner(G_new, self.v[cl, :])
                )
                S_col_new[cl] = C / n_new
                # Row "bmu_temp" - E
                C = (
                    self.CP[cl]
                    + self.n[cl] * np.inner(v_new, v_new)
                    - 2 * np.inner(self.G[cl, :], v_new)
                )
                S_row_new[cl] = C / self.n[cl]

            # Column "ind_minus" - F
            diff_x_v = sample - v_new
            C = (
                self.CP[i_label]
                + np.inner(diff_x_v, diff_x_v)
                + self.n[i_label] * np.inner(v_new, v_new)
                - 2 * np.inner(self.G[i_label, :], v_new)
            )
            S_col_new[i_label] = C / n_new
            S_row_new[i_label] = S_col_new[i_label]

            # Update parameters
            self.n[i_label] = n_new
            self.v[i_label, :] = v_new
            self.CP[i_label] = CP_new
            self.G[i_label, :] = G_new

            # self.S[:, i_label] = S_col_new
            # self.S[i_label, :] = S_row_new
            self.S[i_label, :] = S_col_new
            self.S[:, i_label] = S_row_new

        # Update the parameters that do not depend on label novelty
        self.n_samples = n_samples_new

        return

    @_base.add_docs(_base.param_batch_doc)
    def param_batch(self, data: np.ndarray, labels: np.ndarray) -> None:
        """
        Batch parameter update for the Centroid-based Silhouette (cSIL) CVI.
        """

        # Setup the CVI for batch mode
        super().setup_batch(data)

        # Take the average across all samples, but cast to 1-D vector
        u = np.unique(labels)
        self.n_clusters = u.size
        self.n = np.zeros(self.n_clusters, dtype=int)
        self.v = np.zeros((self.n_clusters, self.dim))
        self.CP = np.zeros(self.n_clusters)
        self.S = np.zeros((self.n_clusters, self.n_clusters))
        D = np.zeros((self.n_samples, self.n_samples))
        for ix in range(self.n_clusters):
            subset_indices = (
                [x for x in range(len(labels)) if labels[x] == ix]
            )
            subset = data[subset_indices, :]
            self.n[ix] = subset.shape[0]
            self.v[ix, :] = np.mean(subset, axis=0)

            # Compute CP in case of switching back to incremental mode
            diff_x_v = subset - self.v[ix, :] * np.ones((self.n[ix], 1))
            self.CP[ix] = np.sum(diff_x_v ** 2)

            d_temp = (data - self.v[ix, :] * np.ones((self.n_samples, 1))) ** 2
            D[ix, :] = np.transpose(np.sum(d_temp, axis=1))
            # D[ix, :] = np.sum(d_temp, axis=1)

        for ix in range(self.n_clusters):
            for jx in range(self.n_clusters):
                subset_ind = [x for x in range(len(labels)) if labels[x] == jx]
                self.S[jx, ix] = sum(D[ix, subset_ind]) / self.n[jx]

        return

    @_base.add_docs(_base.evaluate_doc)
    def evaluate(self) -> None:
        """
        Criterion value evaluation method for the Centroid-based Silhouette (cSIL) CVI.
        """

        self.sil_coefs = np.zeros(self.n_clusters)

        if self.n_clusters > 1 and self.S.any():
            for ix in range(self.n_clusters):
                # Same cluster
                a = self.S[ix, ix]
                # Other clusters
                local_S = np.delete(self.S[:, ix], ix)
                b = np.min(local_S)
                self.sil_coefs[ix] = (b - a) / np.maximum(a, b)
            # cSIL index value
            self.criterion_value = np.sum(self.sil_coefs) / self.n_clusters

        else:
            self.criterion_value = 0.0

        return



"""
Incremental CVI variants.
"""

# Custom imports
import numpy as np


class Cluster:
    def __init__(self,radial=True, box=False):
        self.radial = radial
        self.box = box
        self.len = 0
        self.center = 0
        self.bounding_box = []
        self.points = []
        self.indicies = []
        self.dim = 0

    def add_point(self, x, i=None):
        if not self.points:
            self.dim = len(x)
        assert self.dim == len(x)
        if self.radial:
            new_center = (self.center*self.len + x)/ (self.len+1)
            self.center = new_center
        if self.box:
            assert np.max(x) <= 1
            assert np.min(x) >= 0
            cc_x = np.append(x, 1-x)
            if self.bounding_box:
                self.bounding_box = np.minimum(self.bounding_box, cc_x)
            else:
                self.bounding_box = cc_x

        self.points.append(x)
        if i is not None:
            self.indicies.append(i)
        self.len += 1

    def min_distance(self, x, norm_ord=2):
        min_d = np.inf
        for y in self.points:
            d_y = np.linalg.norm((x-y),norm_ord)
            min_d = np.minimum(min_d,d_y)
        return min_d

    def avg_distance(self,x,norm_ord=2):
        avg_d = 0
        for y in self.points:
            avg_d += np.linalg.norm((x - y), norm_ord)
        avg_d /= self.len
        return avg_d

    def center_distance(self,x,norm_ord=2):
        return np.linalg.norm(self.center-x,norm_ord)

    def within_bounding_box(self,x):
        return np.array_equal(np.minimum(x,self.bounding_box), self.bounding_box)


class Clusters(list):
    def from_list(self, X, C):
        for x, c in zip(X, C):
            if c >= self.__len__():
                self.append(Cluster)
            self[c].add_point(c, c)

def norm22(x):
    return np.linalg.norm(x,2)**2


def CP_update(x,v_old, n_old,g_old=None,cp_old=0,n_new=None,v_new=None,p=2,q=2):
    if g_old is None:
        g_old = np.zeros_like(x)
    if n_new is None:
        n_new = n_old+1
    if v_new is None:
        v_new = v_old + (x-v_old)/n_new
    delta_v = v_old-v_new
    z = x-v_new
    g_new = g_old+z+n_old*delta_v
    cp_new = cp_old+np.linalg.norm(z,q)**p + n_old*np.linalg.norm(delta_v,q)**p + np.sqrt(2*np.dot(delta_v,g_old))**p
    return cp_new, g_new, v_new, n_new

def cluster_center_update(x,v_old,n_old):
    n_new = n_old + 1
    v_new = v_old + (x - v_old) / n_new
    return v_new, n_new



class iXB:
    def __init__(self):
        self.min_v = np.inf
        self.min_v_i = []
        self.N = 0

        self.WGSS = 0
        self.WGSS_i = []
        self.output = 0

        self.cluster_centers = []
        self.cluster_sizes = []
        self.g = None

    def update(self,x,c_i):
        self.N += 1
        if c_i == len(self.cluster_sizes):
            self.cluster_centers.append(x)
            self.cluster_sizes.append(1)
            self.WGSS_i.append(0)
            self.min_v_i.append(np.inf)
        elif c_i > len(self.cluster_sizes):
            raise ValueError('Invalid Cluster Ordering')
        else:
            self.WGSS -= self.WGSS_i[c_i]
            self.WGSS_i[c_i], self.g, self.cluster_centers[c_i],self.cluster_sizes[c_i] = CP_update(x, self.cluster_centers[c_i], self.cluster_sizes[c_i], self.g, self.WGSS_i[c_i])
        # self.WGSS_i[c_i] = CP(self.clusters[c_i].center,self.clusters[c_i],2,2)
        self.WGSS += self.WGSS_i[c_i]
        self.min_v_i[c_i] = np.inf
        for j in range(len(self.cluster_centers)):
            if j != c_i:
                self.min_v_i[c_i] = np.minimum(self.min_v_i[c_i],norm22(self.cluster_centers[c_i]-self.cluster_centers[j]))
        self.min_v = min([self.min_v,self.min_v_i[c_i]])

        self.output = (self.WGSS/self.N)/self.min_v
        return self.output




class iPS:
    def __init__(self):
        self.mean_cluster_center = None
        self.max_cluster_size = 0

        self.output = 0
        self.PS_i = []

        self.cluster_centers = []
        self.cluster_sizes = []

    def update(self,x,c_i):
        if c_i >= len(self.cluster_centers):
            self.cluster_centers.append(x)
            self.cluster_sizes.append(1)
            self.PS_i.append(0)
        else:
            self.cluster_centers[c_i], self.cluster_sizes[c_i] = cluster_center_update(x,self.cluster_centers[c_i],self.cluster_sizes[c_i])
        self.max_cluster_size = np.maximum(self.max_cluster_size, self.cluster_sizes[c_i])
        b_ = [norm22(self.cluster_centers[c_i]-self.cluster_centers[c_j]) for c_j in range(len(self.cluster_centers)) if c_j != c_i]
        if b_:
            b = np.min(b_)
            self.mean_cluster_center = np.sum(self.cluster_centers) / len(self.cluster_centers)
            Bt = np.sum(norm22(cc - self.mean_cluster_center) for cc in self.cluster_centers) / len(self.cluster_centers)
            self.PS_i[c_i] = self.cluster_sizes[c_i] / self.max_cluster_size - np.exp(-b / Bt)
        else:
            self.PS_i[c_i] = 0

            self.output = np.sum(self.PS_i)
        return self.output

class iCH:
    def __init__(self):
        self.clusters = Clusters()

        self.WGSS = 0
        self.WGSS_i = []
        self.BGSS = 0
        self.BGSS_i = []
        self.output = 0
        self.data_center = 0
        self.N = 0

        self.cluster_centers = []
        self.cluster_sizes = []
        self.g = None

    def update(self,x,c_i):
        self.N += 1
        self.data_center = ((self.N - 1) * self.data_center + x) / self.N
        if c_i == len(self.clusters):
            self.cluster_centers.append(x)
            self.cluster_sizes.append(1)
            self.WGSS_i.append(0)
            self.BGSS_i.append(0)
        elif c_i > len(self.cluster_sizes):
            raise ValueError('Invalid Cluster Ordering')
        else:
            self.WGSS -= self.WGSS_i[c_i]
            self.WGSS_i[c_i], self.g, self.cluster_centers[c_i],self.cluster_sizes[c_i] = CP_update(x, self.cluster_centers[c_i], self.cluster_sizes[c_i], self.g, self.WGSS_i[c_i])
            self.WGSS += self.WGSS_i[c_i]

        self.BGSS -= self.BGSS_i[c_i]
        self.BGSS_i[c_i] = self.cluster_sizes[c_i]*norm22(self.cluster_centers[c_i] - self.data_center)
        self.BGSS += self.BGSS_i[c_i]

        if len(self.cluster_centers) > 1 and self.N-len(self.cluster_centers) > 0:
            self.output = (self.BGSS/(len(self.cluster_centers)-1)) / (self.WGSS/(self.N-len(self.cluster_centers)))
        else:
            self.output = 0
        return self.output



class iGD:
    def __init__(self,t=43):

        assert(t==43 or t==53)

        self.t = t
        self.d = np.ones((0,0))*np.inf
        self.D = []
        self.output = 0

        self.cluster_centers = []
        self.cluster_sizes = []
        self.g = None
        self.CP = []

    def update(self,x,c_i):
        if c_i == len(self.cluster_centers):
            self.cluster_centers.append(x)
            self.cluster_sizes.append(1)
            self.CP.append(0)
            self.d = np.pad(self.d, [(0, 1), (0, 1)], mode='constant', constant_values=np.inf)
            self.D.append(-np.inf)
        elif c_i > len(self.cluster_sizes):
            raise ValueError('Invalid Cluster Ordering')
        else:
            self.CP[c_i], self.g, self.cluster_centers[c_i], self.cluster_sizes[c_i] = CP_update(x, self.cluster_centers[c_i], self.cluster_sizes[c_i], self.g, self.CP[c_i], p=1)
            if self.t == 43:
                for j in range(len(self.cluster_centers)):
                    if c_i != j:
                        self.d[c_i,j] = norm22(self.cluster_centers[c_i] - self.cluster_centers[j])
                        self.d[j,c_i] = self.d[c_i,j]
            else:
                for j in range(len(self.cluster_centers)):
                    if c_i != j:
                        self.d[c_i,j] = ( self.CP[c_i] + self.CP[j] ) / (self.cluster_sizes[c_i]+self.cluster_sizes[j])
                        self.d[j,c_i] = self.d[c_i,j]
            self.D[c_i] = 2*self.CP[c_i] / self.cluster_sizes[c_i].len
            self.output = np.min(self.d)/np.max(self.D)

        return self.output

class iGD43(iGD):
    def __init__(self):
        super().__init__(43)

class iGD53(iGD):
    def __init__(self):
        super().__init__(53)

class iSIL:
    def __init__(self):

        self.sc = []
        self.b = np.ones((0,0))*np.inf
        self.output = 0
        self.N = 0

        self.cluster_centers = []
        self.cluster_sizes = []
        self.CP = []
        self.g = []

        self.sij = np.zeros((0,0))

    def cp_update(self,x,i):
        self.CP[i] += np.linalg.norm(x)**2
        self.g[i] += x

    def s_ij_new(self,x,i,j,J,nj,nj_old, cpj_old, vj, vi,gj_old,sij_old):
        zi = x-vi
        zj = x-vj
        if i != J:
            if j == J:
                return (1/nj)*(cpj_old+np.linalg.norm(zi)**2 +nj*np.linalg.norm(vi)**2 - 2*np.dot(vi,gj_old))
            else:
                return sij_old
        else:
            if j == J:
                return (1/nj)*(cpj_old+np.linalg.norm(zj)**2 +nj_old*np.linalg.norm(vj)**2 - 2*np.dot(vj,gj_old))
            else:
                return (1/nj)*(cpj_old +nj*np.linalg.norm(vi)**2 - 2*np.dot(vi,gj_old))

    def s_ij_new_cluster(self,x,i,j,J,nj_old,cpj_old, vi_new, vi_old,gj_old,sij_old):
        if i != J:
            if j == J:
                return np.linalg.norm(x)**2 + np.linalg.norm(vi_old)**2 - 2*np.dot(vi_old,x)
            else:
                return sij_old
        else:
            if j == J:
                return 0
            else:
                return  (1 / nj_old) * (cpj_old + nj_old * np.linalg.norm(vi_new) ** 2 - 2 * np.dot(vi_new, gj_old))

    def sci(self,i, J):
        A = min(self.sij[i,l] for l in range(len(self.cluster_centers)) if l != J) - self.sij[i,J]
        B = max(self.sij[i,J],max(self.sij[i,l] for l in range(len(self.cluster_centers)) if l != J))
        return A/B

    def update(self,x,c_i):
        if c_i == len(self.cluster_sizes):
            self.cluster_centers.append(x)
            self.cluster_sizes.append(1)
            self.sc.append(0)
            self.b = np.pad(self.b, [(0, 1), (0, 1)], mode='constant', constant_values=np.inf)
            self.sij = np.pad(self.b, [(0, 1), (0, 1)], mode='constant', constant_values=0)
            self.CP.append(np.linalg.norm(x) ** 2)
            self.g.append(x)
        elif c_i > len(self.cluster_sizes):
            raise ValueError('Invalid Cluster Ordering')
        self.N += 1
        self.cluster_centers[c_i], self.cluster_sizes[c_i] = cluster_center_update(x, self.cluster_centers[c_i], self.cluster_sizes[c_i])
        for i in range(len(self.cluster_sizes)):
            for j in range(len(self.cluster_sizes)):
                self.sij[i, j] = self.s_ij_new(x, i, j, c_i, self.cluster_sizes[j], self.cluster_sizes[j]-1, self.CP[j], self.cluster_centers[j], self.cluster_centers[i], self.g[j], self.sij[i, j])
        self.cp_update(x, c_i)
        self.output = (1/len(self.cluster_centers))*sum([self.sci(i,c_i) for i in range(len(self.cluster_centers))])
        return self.output

class iDB:
    def __init__(self):

        self.R = []
        self.S = []
        self.M = np.zeros((0,0))
        self.p = 1
        self.q = 1
        self.output = 0
        self.N = 0

        self.cluster_centers = []
        self.cluster_sizes = []
        self.g = None
        self.CP = []

    def update(self,x,c_i):
        self.N += 1
        if c_i == len(self.cluster_centers):
            self.cluster_centers.append(x)
            self.cluster_sizes.append(1)
            self.R.append(0)
            self.S.append(0)
            self.M = np.pad(self.M, [(0, 1), (0, 1)], mode='constant', constant_values=0)
        elif c_i > len(self.cluster_sizes):
            raise ValueError('Invalid Cluster Ordering')
        else:
            self.CP[c_i], self.g, self.cluster_centers[c_i], self.cluster_sizes[c_i] = CP_update(x,self.cluster_centers[c_i],self.cluster_sizes[c_i], self.g,self.CP[c_i], p=self.p, q=self.q)
            self.S[c_i] = ((1/self.cluster_sizes[c_i])*self.CP[c_i])**(1/self.q)
            self.R[c_i] = 0
            for j in range(len(self.cluster_sizes)):
                if j != c_i:
                    self.M[c_i,j] = 0
                    for t in range(len(x)):
                        if self.cluster_sizes[j] > 0:
                            self.M[c_i,j] = abs(self.cluster_centers[c_i][t]-self.cluster_centers[j][t])**self.p
                    self.M[c_i,j] = self.M[c_i,j]**(1./self.p)
                    self.M[j,c_i] = self.M[c_i,j]
                    if self.cluster_sizes[j] > 0:
                        self.R[c_i] = max(self.R[c_i], (self.S[c_i]+self.S[j])/self.M[c_i,j])

            self.output = np.sum(self.R)/sum([clen > 0 for clen in self.cluster_sizes])
        return self.output

def iCVI(name):
    if name == 'iDB':
        return iDB()
    elif name == 'iSIL':
        return iSIL()
    elif name == 'GD43':
        return iGD43()
    elif name == 'iGD53':
        return iGD53()
    elif name == 'iCH':
        return iCH()
    elif name == 'iPS':
        return iPS()
    elif name == 'iXB':
        return iXB()
    else:
        raise ValueError('iCVI {} not implemented'.format(name))



"""
Utilities that are common across all CVI objects.
"""

# Standard library imports
from typing import (
    Callable,
    Union
)
from abc import abstractmethod

# Custom imports
import numpy as np

# --------------------------------------------------------------------------- #
# CLASSES
# --------------------------------------------------------------------------- #


class LabelMap():
    """
    Internal map between labels and the incremental CVI categories.
    """

    def __init__(self) -> None:
        self.map = dict()
        return

    def get_internal_label(self, label: int) -> int:
        """
        Gets the internal label and updates the label map if the label is new.
        """

        # Initialize the internal label
        internal_label = None

        # If the label is in the map, return that
        if label in self.map:
            internal_label = self.map[label]
        # Otherwise, create an incremented new label and return that
        else:
            # Correct for python zero-indexing by not including the +1
            internal_label = len(self.map.items())
            self.map[label] = internal_label

        return internal_label


class CVI():
    """
    Superclass containing elements shared between all CVIs.
    """

    def __init__(self) -> None:
        """
        CVI base class initialization method.
        """
        self.label_map = LabelMap()
        self.dim = 0
        self.n_samples = 0
        self.n = []                 # dim
        self.v = np.zeros([0, 0])   # n_clusters x dim
        self.CP = []                # dim
        self.G = np.zeros([0, 0])   # n_clusters x dim
        self.n_clusters = 0
        self.criterion_value = 0.0
        self.is_setup = False

        return

    def setup(self, sample: np.ndarray) -> None:
        """
        Common CVI procedure for incremental setup.

        Parameters
        ----------
        data : numpy.ndarray
            Sample vector of features.
        """

        # Infer the dimension as the length of the provided sample
        self.dim = len(sample)

        # Set the sizes of common arrays for consistent appending
        self.v = np.zeros([0, self.dim])
        self.G = np.zeros([0, self.dim])

        # Declare that the CVI is internally setup
        self.is_setup = True

        return

    def setup_batch(self, data: np.ndarray) -> None:
        """
        Common CVI procedure for batch setup.

        Parameters
        ----------
        data : np.ndarray
            A batch of samples with some feature dimension.
        """

        # Infer the data dimension and number of samples
        self.n_samples, self.dim = data.shape
        self.is_setup = True

        return

    @abstractmethod
    def param_inc(self, sample: np.ndarray, label: int) -> None:
        pass

    @abstractmethod
    def param_batch(self, data: np.ndarray, labels: np.ndarray) -> None:
        pass

    @abstractmethod
    def evaluate(self) -> None:
        pass

    def get_cvi(self, data: np.ndarray, label: Union[int, np.ndarray]) -> float:
        """
        Updates the CVI parameters and then evaluates and returns the criterion value.

        Parameters
        ----------
        data : np.ndarray
            The sample(s) of features used for clustering.
        label : Union[int, np.ndarray]
            The label(s) prescribed to the sample(s) by the clustering algorithm.

        Returns
        -------
        float
            The CVI's criterion value.
        """

        # If we got 1D data, do a quick update
        if (data.ndim == 1):
            self.param_inc(data, label)
            pass
        # Otherwise, we got 2D data and do the correct update
        elif (data.ndim == 2):
            # If we haven't done a batch update yet
            if not self.is_setup:
                # Do a batch update
                self.param_batch(data, label)
            # Otherwise, we are already setup
            else:
                raise ValueError(
                    "Switching from batch to incremental not supported"
                )
                # Do many incremental updates
                # for ix in range(len(label)):
                #     self.param_inc(data[ix, :], label[ix])
        else:
            raise ValueError(
                f"Please provide 1D or 2D numpy array, recieved ndim={data.ndim}"
            )

        # Regardless of path, evaluate and extract the criterion value
        self.evaluate()
        criterion_value = self.criterion_value

        # Return the criterion value
        return criterion_value


# --------------------------------------------------------------------------- #
# DECORATORS
# --------------------------------------------------------------------------- #


def add_docs(other_func: Callable[[], None]) -> Callable[[], None]:
    """
    A decorator for appending the docstring of one function to another.

    Parameters
    ----------
    other_func : Callable[[], None]
        The other function whose docstring you want to append to the decorated function.
    """

    def dec(func):
        func.__doc__ = func.__doc__ + other_func.__doc__
        return func

    return dec


# --------------------------------------------------------------------------- #
# DOCSTRING FUNCTIONS
# --------------------------------------------------------------------------- #


def setup_doc() -> None:
    """
    Sets up the dimensions of the CVI based on the sample size.

    Parameters
    ----------
    sample : numpy.ndarray
        A sample vector of features.
    """

    pass


# This function documents the shared API for incremental parameter updates
def param_inc_doc() -> None:
    """
    Parameters
    ----------
    sample : numpy.ndarray
        A sample row vector of features.
    label : int
        An integer label for the cluster, zero-indexed.
    """

    pass


# This function documents the shared API for batch parameter updates
def param_batch_doc() -> None:
    """
    Parameters
    ----------
    sample : numpy.ndarray
        A batch of samples; each row is a new sample of features.
    label : numpy.ndarray
        A vector of integer labels, zero-indexed.
    """

    pass


# This function documents the shared API for criterion value evaluation
def evaluate_doc() -> None:
    """
    Updates the internal `criterion_value` parameter.
    """

    pass



"""
The Calinski-Harabasz (CH) Cluster Validity Index.
"""

# Custom imports
import numpy as np

# Local imports
from . import _base


# CH object definition
class CH(_base.CVI):
    """
    The stateful information of the Calinski-Harabasz (CH) Cluster Validity Index.

    References
    ----------
    1. L. E. Brito da Silva, N. M. Melton, and D. C. Wunsch II, "Incremental Cluster Validity Indices for Hard Partitions: Extensions  and  Comparative Study," ArXiv  e-prints, Feb 2019, arXiv:1902.06711v1 [cs.LG].
    2. T. Calinski and J. Harabasz, "A dendrite method for cluster analysis," Communications in Statistics, vol. 3, no. 1, pp. 1-27, 1974.
    3. M. Moshtaghi, J. C. Bezdek, S. M. Erfani, C. Leckie, and J. Bailey, "Online Cluster Validity Indices for Streaming Data," ArXiv e-prints, 2018, arXiv:1801.02937v1 [stat.ML]. [Online].
    4. M. Moshtaghi, J. C. Bezdek, S. M. Erfani, C. Leckie, J. Bailey, "Online cluster validity indices for performance monitoring of streaming data clustering," Int. J. Intell. Syst., pp. 1-23, 2018.
    """

    def __init__(self):
        """
        CH initialization routine.
        """

        # Run the base initialization
        super().__init__()

        # CH-specific initialization
        self.mu = np.zeros([0])     # dim
        self.SEP = np.zeros([0])     # dim
        self.BGSS = 0.0
        self.WGSS = 0.0

        return

    @_base.add_docs(_base.setup_doc)
    def setup(self, sample: np.ndarray) -> None:
        """
        Calinski-Harabasz (CH) setup routine.
        """

        # Run the generic setup routine
        super().setup(sample)

        # CH-specific setup
        self.SEP = np.zeros([self.dim])
        self.mu = sample

        return

    @_base.add_docs(_base.param_inc_doc)
    def param_inc(self, sample: np.ndarray, label: int) -> None:
        """
        Incremental parameter update for the Calinski-Harabasz (CH) CVI.
        """

        # Get the internal label corresponding to the provided label
        i_label = self.label_map.get_internal_label(label)

        # Increment the local number of samples count
        n_samples_new = self.n_samples + 1

        # Check if the module has been setup, then set the mu accordingly
        if self.n_samples == 0:
            self.setup(sample)
        else:
            self.mu = (1 - 1/n_samples_new) * self.mu + (1/n_samples_new) * sample

        # IF NEW CLUSTER LABEL
        # Correct for python 0-indexing
        if i_label > self.n_clusters - 1:
            n_new = 1
            v_new = sample
            CP_new = 0.0
            G_new = np.zeros(self.dim)

            # Update 1-D parameters with list appends
            self.n_clusters += 1
            self.n.append(n_new)
            self.CP.append(CP_new)

            # Update 2-D parameters with numpy vstacks
            self.v = np.vstack([self.v, v_new])
            self.G = np.vstack([self.G, G_new])

        # ELSE OLD CLUSTER LABEL
        else:
            n_new = self.n[i_label] + 1
            v_new = (
                (1 - 1 / n_new) * self.v[i_label, :]
                + (1 / n_new) * sample
            )
            delta_v = self.v[i_label, :] - v_new
            diff_x_v = sample - v_new
            CP_new = (
                self.CP[i_label]
                + np.inner(diff_x_v, diff_x_v)
                + self.n[i_label] * np.inner(delta_v, delta_v)
                + 2 * np.inner(delta_v, self.G[i_label, :])
            )
            G_new = (
                self.G[i_label, :]
                + diff_x_v
                + self.n[i_label] * delta_v
            )
            # Update parameters
            self.n[i_label] = n_new
            self.v[i_label, :] = v_new
            self.CP[i_label] = CP_new
            self.G[i_label, :] = G_new

        # Update the parameters that do not depend on label novelty
        self.n_samples = n_samples_new
        # self.mu = mu_new
        self.SEP = np.array([self.n[ix] * sum((self.v[ix, :] - self.mu)**2) for ix in range(self.n_clusters)])

        return

    @_base.add_docs(_base.param_batch_doc)
    def param_batch(self, data: np.ndarray, labels: np.ndarray) -> None:
        """
        Batch parameter update for the Calinski-Harabasz (CH) CVI.
        """

        # Setup the CVI for batch mode
        super().setup_batch(data)

        # Take the average across all samples, but cast to 1-D vector
        self.mu = np.mean(data, axis=0)
        u = np.unique(labels)
        self.n_clusters = u.size
        self.n = np.zeros(self.n_clusters, dtype=int)
        self.v = np.zeros((self.n_clusters, self.dim))
        self.CP = np.zeros(self.n_clusters)
        self.SEP = np.zeros(self.n_clusters)

        for ix in range(self.n_clusters):
            subset_indices = (
                [x for x in range(len(labels)) if labels[x] == ix]
            )
            subset = data[subset_indices, :]
            self.n[ix] = subset.shape[0]
            self.v[ix, :] = np.mean(subset, axis=0)
            diff_x_v = subset - self.v[ix, :] * np.ones((self.n[ix], 1))
            self.CP[ix] = np.sum(diff_x_v ** 2)
            self.SEP[ix] = self.n[ix] * np.sum((self.v[ix, :] - self.mu) ** 2)

        return

    @_base.add_docs(_base.evaluate_doc)
    def evaluate(self) -> None:
        """
        Criterion value evaluation method for the Calinski-Harabasz (CH) CVI.
        """

        if self.n_clusters > 2:
            # Within group sum of scatters
            self.WGSS = sum(self.CP)
            # Between groups sum of scatters
            self.BGSS = sum(self.SEP)
            # CH index value
            self.criterion_value = (
                (self.BGSS / self.WGSS)
                * ((self.n_samples - self.n_clusters) / (self.n_clusters - 1))
            )
        else:
            self.BGSS = 0.0
            self.criterion_value = 0.0

        return



"""
cvi - A Python library for both incremental and batch cluster validity indices.
"""

__version__ = "0.1.0"

from .modules import (
    CVI,
    CH,
    cSIL,
    DB,
)

__all__ = [
    "CVI",
    "CH",
    "cSIL",
    "DB",
]
